/*
http://homes.cs.washington.edu/~djg/teachingMaterials/spac/sophomoricParallelismAndConcurrency.pdf#page=18
*/

import java.lang.Thread;
import java.util.Random;

/**
   // Sum an array of integrs.
   int sum(int[] arr)
   {
      int ans = 0;
      for (int i = 0; i < arr.length; i++)
      {
         ans += arr[i];
      }
      return ans;
   }

   // Pseudocode for summing an array of integers in parallel.
   int sum(int[] arr)
   {
      results = new int[4];
      len = arr.length;
      FORALL (i = 0; i < 4; ++i)
      {
         results[i] = sumRange(arr,(i*len)/4,((i+1)*len)/4);
      }
      return results[0] + results[1] + results[2] + results[3];
   }

   int sumRange(int[] arr, int lo, int hi)
   {
      result = 0;
      for (j = lo; j < hi; ++j)
      {
         result += arr[j];
      }
      return result;
   }
*/

class SumThread extends Thread
{
   int lo;  // fields for communicating inputs
   int hi;
   int[] arr;
   long ans = 0; // for communicating result

   SumThread(int[] a, int l, int h)
   {
      arr=a; lo=l; hi=h;
   }

   public void run()  // overriding, must have this type
   {
      for (int i = lo; i < hi; i++)
      {
         ans += arr[i];
      }
   }
}


class C
{
   static long sum(int[] arr) throws java.lang.InterruptedException
   {
      int len = arr.length;
      long ans = 0;
      SumThread[] ts = new SumThread[4];

      for (int i = 0; i < 4; i++)
      {
         ts[i] = new SumThread(arr,(i*len)/4,((i+1)*len)/4);
         ts[i].start();
      }

      for (int i = 0; i < 4; i++)
      {
         ts[i].join();
         ans += ts[i].ans;
      }
      return ans;
   }
}


public class Section_3_1
{
   public static final int PROBLEM_SIZE = 100_000_000;
   public static int numbers[] = new int[PROBLEM_SIZE];

   public static void main(String[] args) throws java.lang.InterruptedException
   {
      Random generator = new Random();
      for (int i = 0; i < numbers.length; i++)
         numbers[i] = generator.nextInt(40);

      long startTime;
      long stopTime;

      startTime = System.currentTimeMillis();
         long result = C.sum(numbers);
      stopTime = System.currentTimeMillis();
      System.out.println("Parallel sum: " + result);
      System.out.println("Parallel solution time: " + (stopTime - startTime));

      result = 0;
      startTime = System.currentTimeMillis();
        for (int i = 0; i < numbers.length; ++i)
           result += numbers[i];
      stopTime = System.currentTimeMillis();
      System.out.println("Sequential sum: " + result);
      System.out.println("Sequential solution time: " + (stopTime - startTime));
   }
}
