/*
http://homes.cs.washington.edu/~djg/teachingMaterials/spac/sophomoricParallelismAndConcurrency.pdf#page=24
*/

import java.lang.Thread;
import java.util.Random;

class SumThread extends Thread
{
   static int SEQUENTIAL_CUTOFF = 1_000;

   int lo;  // fields for communicating inputs
   int hi;
   int[] arr;
   long ans = 0; // for communicating result

   SumThread(int[] a, int l, int h)
   {
      arr=a; lo=l; hi=h;
   }

   public void run()  // overriding, must have this type
   {
      if (hi - lo <= SEQUENTIAL_CUTOFF)
      {
         for (int i = lo; i < hi; i++)
            ans += arr[i];
      }
      else
      {
         SumThread left  = new SumThread(arr, lo, (hi+lo)/2);
         SumThread right = new SumThread(arr, (hi+lo)/2, hi);

         left.start();   // asynchronous thread
         right.start();  // asynchronous thread
         try
         {
            left.join();
            right.join();
         }
         catch (java.lang.InterruptedException e){}
/*
         left.start();   // asynchronous thread
         right.run();    // synchronous function call
         try
         {
            left.join();
         }
         catch (java.lang.InterruptedException e){}
*/
         ans = left.ans + right.ans;
      }
   }
}


class C
{
   static long sum(int[] arr)
   {
      SumThread t = new SumThread(arr, 0, arr.length);
      t.run();     // synchronous function call
      return t.ans;
   }
}


public class Section_3_3
{
   public static final int PROBLEM_SIZE = 100_000_000;
   public static int numbers[] = new int[PROBLEM_SIZE];

   public static void main(String[] args)
   {
      Random generator = new Random();
      for (int i = 0; i < numbers.length; i++)
         numbers[i] = generator.nextInt(40);

      long startTime;
      long stopTime;

      startTime = System.currentTimeMillis();
         long result = C.sum(numbers);
      stopTime = System.currentTimeMillis();
      System.out.println("Parallel sum: " + result);
      System.out.println("Parallel solution time: " + (stopTime - startTime));

      result = 0;
      startTime = System.currentTimeMillis();
        for (int i = 0; i < numbers.length; ++i)
           result += numbers[i];
      stopTime = System.currentTimeMillis();
      System.out.println("Sequential sum: " + result);
      System.out.println("Sequential solution time: " + (stopTime - startTime));
   }
}
