/*
http://homes.cs.washington.edu/~djg/teachingMaterials/spac/sophomoricParallelismAndConcurrency.pdf#page=27
*/

import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveTask;
import java.util.Random;

@SuppressWarnings("serial")
class SumArray extends RecursiveTask<Long>
{
   static int SEQUENTIAL_CUTOFF = 1_000;

   int lo;
   int hi;
   int[] arr;


   SumArray(int[] a, int l, int h)
   {
      arr=a; lo=l; hi=h;
   }

   public Long compute()
   {
      if (hi - lo <= SEQUENTIAL_CUTOFF)
      {
         long ans = 0;
         for (int i=lo; i < hi; ++i)
            ans += arr[i];
         return ans;
      }
      else
      {
         SumArray left  = new SumArray(arr,lo,(hi+lo)/2);
         SumArray right = new SumArray(arr,(hi+lo)/2,hi);

         left.fork();                     // asynchronous, put left item in the work queue
         long rightAns = right.compute(); // synchronous, continue working on the right item
         long  leftAns = left.join();     // whenever some worker thread completes left item
         return leftAns + rightAns;
      }
   }
}


class C
{
   static long sum(int[] array)
   {
      // Create the initial task item.
      SumArray t = new SumArray(array, 0, array.length);

      // Put item t into the fork/join work queue.
      return ForkJoinPool.commonPool().invoke(t);
   }
}


public class Section_3_4b
{
   public static final int PROBLEM_SIZE = 100_000_000;
   public static int numbers[] = new int[PROBLEM_SIZE];

   public static void main(String[] args)
   {
      Random generator = new Random();
      for (int i = 0; i < numbers.length; i++)
         numbers[i] = generator.nextInt(40);

      long startTime;
      long stopTime;

      SumArray.SEQUENTIAL_CUTOFF = 5000;  // BE SURE TO EXPERIMENT WITH THIS!
      startTime = System.currentTimeMillis();
         long result = C.sum(numbers);
      stopTime = System.currentTimeMillis();
      System.out.println(" Fork-join sum: " + result);
      System.out.println(" Fork-join solution time: " + (stopTime - startTime));
      System.out.println("ForkJoinPool: " + ForkJoinPool.commonPool().toString());

      result = 0;
      startTime = System.currentTimeMillis();
      for (int i = 0; i < numbers.length; ++i)
         result += numbers[i];
      stopTime = System.currentTimeMillis();
      System.out.println(" Squential sum: " + result);
      System.out.println(" Sequential solution time: " + (stopTime - startTime));
   }
}
