/*
    Form "Java 7 New Features Cookbook"
    Chapter 10, pages 300-301.
*/

import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveTask;

@SuppressWarnings("serial")
class SumOfSquaresTask extends RecursiveTask<Long>
{
   private final int THRESHOLD = 1_000; /*** BE SURE TO EXPERIMENT WITH THIS! ***/
   private final int[] numbers;
   private final int from;
   private final int to;

   public SumOfSquaresTask(int[] numbers, int from, int to)
   {
      this.numbers = numbers;
      this.from = from;
      this.to = to;
   }


   @Override
   protected Long compute()
   {
      final int mid = (to + from) >>> 1;
      long sum = 0L;
      if ((to - from) < THRESHOLD)
      {
         for (int i = from; i < to; i++)
         {
            sum += numbers[i] * numbers[i];
         }
         return sum;
      }
      else
      {
         SumOfSquaresTask task1 = new SumOfSquaresTask(numbers, from, mid);
         SumOfSquaresTask task2 = new SumOfSquaresTask(numbers, mid, to);

         task1.fork();
         task2.fork();

         sum += task1.join();
         sum += task2.join();

         return sum;
      }
   }
}


public class SumOfSquares
{
   public static final int PROBLEM_SIZE = 300_000_000;
   public static final int numbers[] = new int[PROBLEM_SIZE];

   public static void main(String[] args)
   {
      System.out.println("Generate array of integers from 1 to " + PROBLEM_SIZE + ".");
      for (int i = 0; i < numbers.length; ++i)
      {
         numbers[i] = i;
      }

      long startTime;
      long stopTime;
      long sum = 0L;

      // Run a single-threaded sequential solution.
      startTime = System.currentTimeMillis();
      for (int i = 0; i < numbers.length; ++i)
      {
         sum += numbers[i] * numbers[i];
      }
      stopTime = System.currentTimeMillis();
      System.out.println("Iterative sum of squares: " + sum);
      System.out.println("Iterative solution time: " + (stopTime - startTime));

      // Run the fork-join version.
      startTime = System.currentTimeMillis();
        final long result = ForkJoinPool.commonPool().invoke(new SumOfSquaresTask(numbers, 0, numbers.length));
      stopTime = System.currentTimeMillis();
      System.out.println("Fork-join sum of squares: " + result);
      System.out.println("Fork-join solution time: " + (stopTime - startTime));
      System.out.println("ForkJoinPool: " + ForkJoinPool.commonPool().toString());
   }
}
